#include "bonus_points.hpp"

#include "util/enums.hpp"
#include "util/exception.hpp"
#include "util/heuristics.hpp"
#include "util/points.hpp"
#include "util/strings.hpp"

static void throwIfWrongFileType(const QString &firstCell) noexcept(false) {
    const auto fileType = getApproximateFileType(firstCell);

    switch (fileType) {
    case FileType::dynexite:
        throw BonusParseException(str::error::unexpectedDynexite);
    case FileType::rwth:
        throw BonusParseException(str::error::unexpectedRwthOnline);
    default:
        // the file might still be invalid, but that will be decided later
        break;
    }
}

Points
impl::getNrBonusPoints(const QList<QPair<QString, Points>> &bonusPointList,
                       const QString &regNr) noexcept {
    for (const auto &[pairRegNr, pairBonusPoints] : bonusPointList) {
        if (pairRegNr == regNr) {
            return pairBonusPoints;
        }
    }
    return Points{0.};
}

Points BonusPoints::getNrBonusPoints(const QString &regNr) const noexcept {
    return impl::getNrBonusPoints(mTotalBonusPointsPerStudent, regNr);
}

BonusPoints::BonusPoints(const CsvDump &c) noexcept(false) : mBonusPoints(c) {

    // empty bonus point file is ok, just no bonus points for anybody
    if (c.data.empty()) {
        return;
    }

    // If we do have data, we need a header
    if (c.data[0].empty())
        throw BonusParseException(str::error::noHeader);

    // Quick check: reading wrong file type?
    throwIfWrongFileType(c.data[0][0]);

    const auto nrRows = c.data.size();

    // No data aside from header is allowed
    if (nrRows < 2)
        return;

    mTotalBonusPointsPerStudent.reserve(nrRows - 1);

    // Get column index
    const auto registrationNrColumn = c.data[0].indexOf("Matrikelnummer");
    const auto bonusPointColumn = c.data[0].indexOf("Bonuspunkte (Punkte)");

    // Throw if relevant columns not found
    if (registrationNrColumn == -1)
        throw BonusParseException(str::error::noIdColumn);
    if (bonusPointColumn == -1)
        throw BonusParseException(str::error::noBonusPointColumn);

    // Insert relevant data into lists
    for (qsizetype i = 1; i < nrRows; ++i) {

        // Registration number column sometimes contains an email address
        // instead of a registration number. Issue a warning for all those
        // students.
        // If the bonus point number can't be read, issue an warning and don't
        // count the bonus points.

        bool regNrParseSuccess, bonusPointParseSuccess;
        c.data[i][registrationNrColumn].toInt(&regNrParseSuccess);

        Points bonusPoints;
        try {
            bonusPoints = stringToPoints(c.data[i][bonusPointColumn]);
            bonusPointParseSuccess = true;
        } catch (...) {
            bonusPointParseSuccess = false;
        }

        if (bonusPointParseSuccess) {
            mTotalBonusPointsPerStudent.emplace_back(
                c.data[i][registrationNrColumn], bonusPoints);
        }

        if (!bonusPointParseSuccess || !regNrParseSuccess) {
            mInvalidRegNrs.emplace_back(c.data[i][registrationNrColumn],
                                        c.data[i][bonusPointColumn]);
        }
    }
}
