#pragma once

#include "node.hpp"

#include "data/csv_dump.hpp"
#include "util/points.hpp"
#include "util/storage_helper.hpp"

#include <QList>
#include <QString>

/**
 * \brief Directly represents the data in the csv-file directly from dynexite in
 * a more accessible manner. Does not change after initialization.
 *
 * Stores data only for students who attended, discards everyone else.
 *
 * \note Data layer 1.
 */
struct DynexiteExam : Node<> {
    using Node<>::Node;

    /**
     * \brief Interprets the CsvDump as a dynexite-csv-file and copies all
     * necessary info. Also runs sanity checks on the given data and throws on
     * error.
     *
     * \exception std::invalid_argument A sanity check on the given csv file
     * failed, see the description in the exception for details, or a std::stod
     * conversion failed.
     * \exception std::out_of_range Thrown by std::stod.
     */
    explicit DynexiteExam(const CsvDump &c) noexcept(false);

    /**"
     * \brief Unique identifier (usually Matrikelnummer), called mRegistrationNr
     * in the students data
     */
    QList<QString> mIdentifier{};

    /**
     * \brief The maximum achievable open, closed and combined points as stated
     * by dynexite. Not influenced by later changes from the user, those are
     * stored in \ref ChangedMaxPoints
     */
    EnumStorage<Points> mPointsMax{};

    /**
     * \brief Open, closed and combined achieved points of each student, without
     * bonus points.
     */
    EnumStorage<QList<Points>> mPointsWithoutBonus{};

    /**
     * \brief For each task holds the earned open, closed and combined points
     * each student got in that task, if it was assigned to them
     */
    EnumStorage<QList<QList<Points>>> mTaskEarnedPoints{};

    /**
     * \brief Holds the maximal (open+closed) points for each task.
     */
    QList<Points> mTaskMaxPoints{};

    /**
     * \brief Holds the name of each task.
     */
    QList<QString> mTaskNames{};

    // Store the entire csv dump so that we can more easily export it later
    CsvDump mDynexite;

private:
    // internal helper functions
    void reserveArrays(qsizetype nrStudents, qsizetype nrTasks) noexcept;
    bool insertData(const CsvDump &c,
                    CsvLine::size_type nrTasks) noexcept(false);
};
