#include "grade_list.hpp"

#include "dynexite_exam.hpp"
#include "user_point_config.hpp"

#include <algorithm>
#include <cassert>

/**
 * \brief Computes a grade for the given points with the given grade boundaries
 */
template <GradingType Grading>
[[nodiscard]] static Grade
mapPointsToGrade(const Points p,
                 const UserGradeBoundaries<Grading> &gradeBoundaries,
                 const GradeBoundaryStepSize stepSize) noexcept {

    // Boundaries are sorted from good to bad grades, so from many to few
    // points. Find the first boundary b that does not satisfy b > points.
    const auto res = std::lower_bound(
        gradeBoundaries.mBoundaries.begin(), gradeBoundaries.mBoundaries.end(),
        p, [stepSize](const GradeBoundary &b, const Points &points) {
            return boundaryGreater(b, points, stepSize);
        });

    if (res != gradeBoundaries.mBoundaries.end()) {

        // res is now an iterator to the grade boundary of the resulting grade
        return static_cast<Grade>(res - gradeBoundaries.mBoundaries.begin());

    } else {
        // Since gradeBoundaries.mBoundaries[Grade::grade5_0] = 0, this
        // should be unreachable
        assert(false);
        return Grade::grade5_0;
    }
}

/**
 * \brief Computes a grade for the given points and bonus points with the given
 * grade boundaries
 */
template <GradingType Grading>
[[nodiscard]] static Grade
getGrade(const Points pointsWithoutBonus, const Points pointsWithBonus,
         const UserGradeBoundaries<Grading> &gradeBoundaries,
         const GradeBoundaryStepSize stepSize) noexcept {

    // First calculate the grade without taking bonus points into account
    const auto gradeWithoutBonus =
        mapPointsToGrade(pointsWithoutBonus, gradeBoundaries, stepSize);

    // Bonus points are only taken into account if the student passes the exam
    // without bonus points
    if (gradeWithoutBonus == Grade::grade5_0 ||
        pointsWithoutBonus == pointsWithBonus) {
        return gradeWithoutBonus;
    }

    // If the student passes without bonus points, calculate the final grade
    // with bonus points
    return mapPointsToGrade(pointsWithBonus, gradeBoundaries, stepSize);
}

template <GradingType Grading>
void GradeList<Grading>::recalculate() {

    const auto &userGradeBoundsRef =
        std::get<const UserGradeBoundaries<Grading> &>(this->mParentRefs);
    const auto &examRef = std::get<const DynexiteExam &>(this->mParentRefs);
    const auto &studentPointsRef =
        std::get<const StudentPoints<Grading> &>(this->mParentRefs);
    const auto &configRef =
        std::get<const UserPointConfig &>(this->mParentRefs);

    // Resize the QList, reset the grades
    mGrades = QList<Grade>(examRef.mPointsWithoutBonus[Grading].size());

    // Generate a list of grades in the same order the points were in
    for (qsizetype i = 0; i < mGrades.size(); ++i) {

        mGrades[i] =
            getGrade(examRef.mPointsWithoutBonus[Grading][i],
                     studentPointsRef.mPointsWithBonus[i], userGradeBoundsRef,
                     configRef.mGradeBoundaryStepSize);
    }
}

template struct GradeList<GradingType::Closed>;
template struct GradeList<GradingType::Open>;
