#include "student_data.hpp"

#include "util/exception.hpp"
#include "util/heuristics.hpp"
#include "util/strings.hpp"

#include <QPair>

#include <array>
#include <ranges>
#include <string>

#define RWTHONLINE_PARSER_PERMISSIVE 1

[[maybe_unused]] static void checkHeader(const QString &found,
                                         const QString &expected) {
    if (found != expected) {
        throw RwthOnlineParseException("Header validation error: expected \"" +
                                       expected.toStdString() + "\", found \"" +
                                       found.toStdString() + "\"");
    }
}

/**
 * \brief Check the naming of all headers.
 */
static QPair<qsizetype, qsizetype> validateHeaders(const CsvDump &c) {

    // Do headers exist at all?
    if (c.data.empty())
        throw RwthOnlineParseException(str::error::empty);

    const auto &headers = c.data[0];

    // In permissive mode we only check for the names of headers we require and
    // ignore all other header names.
#if !RWTHONLINE_PARSER_PERMISSIVE
    if (c.data[0].size() != 24) {
        throw RwthOnlineParseException("Header validation: "
                                       "Unexpected number of columns: " +
                                       std::to_string(c.data[0].size()) +
                                       " (To disable this check, set "
                                       "RWTHONLINE_PARSER_PERMISSIVE to true)");
    }

    static constexpr std::array headerStrings = {"STUDY_PROGRAMME",
                                                 "CODE_OF_STUDY_PROGRAMME",
                                                 "Studienplan_Version",
                                                 "SPO_KONTEXT",
                                                 "REGISTRATION_NUMBER",
                                                 "FAMILY_NAME_OF_STUDENT",
                                                 "FIRST_NAME_OF_STUDENT",
                                                 "GESCHLECHT",
                                                 "DATE_OF_ASSESSMENT",
                                                 "GUEL_U_AKTUELLE_ANTRITTE_SPO",
                                                 "GRADE",
                                                 "REMARK",
                                                 "Number_Of_The_Course",
                                                 "SEMESTER_OF_The_COURSE",
                                                 "COURSE_TITLE",
                                                 "Examiner",
                                                 "Start_Time",
                                                 "TERMIN_ORT",
                                                 "DB_Primary_Key_Of_Exam",
                                                 "DB_Primary_Key_Of_Candidate",
                                                 "COURSE_GROUP_NAME",
                                                 "FILE_REMARK",
                                                 "EMAIL_ADDRESS",
                                                 "ECTS_GRADE"};

    for (int i = 0; i < 24; ++i) {
        checkHeader(headers[i], headerStrings[i]);
    }
    return {4u, 9u};
#else
    // find column of registration number and attempt nr
    const auto registrationNrIdx = headers.indexOf("REGISTRATION_NUMBER");

    const auto attemptNrIdx = headers.indexOf("GUEL_U_AKTUELLE_ANTRITTE_SPO");

    if (registrationNrIdx == -1) {
        throw RwthOnlineParseException(str::error::noIdColumn);
    }
    if (attemptNrIdx == -1) {
        throw RwthOnlineParseException(str::error::noAttemptNrColumn);
    }
    return {registrationNrIdx, attemptNrIdx};
#endif
}

StudentData::StudentData(const CsvDump &c) noexcept(false) : mRwthOnline(c) {
    if (c.data.empty())
        throw RwthOnlineParseException(str::error::empty);

    if (c.data[0].empty())
        throw RwthOnlineParseException(str::error::noHeader);

    // Quick check: reading wrong file type?
    {
        const auto fileType = getApproximateFileType(c.data[0][0]);

        switch (fileType) {
        case FileType::bonus:
            throw RwthOnlineParseException(str::error::unexpectedBonus);
        case FileType::dynexite:
            throw RwthOnlineParseException(str::error::unexpectedDynexite);
        default:
            // the file might still be invalid, but that will be decided later
            break;
        }
    }

    const auto nrRows = c.data.size();

    // No data aside from header?
    if (nrRows < 2)
        throw RwthOnlineParseException(str::error::noData);

    const auto nrStudents = nrRows - 1;
    reserveArrays(nrStudents);

    const auto [regIdx, attemptIdx] = validateHeaders(c);

    insertData(c, regIdx, attemptIdx);
}

void StudentData::reserveArrays(const qsizetype nrStudents) noexcept {
    mRegistrationNr.reserve(nrStudents);
    mAttemptNr.reserve(nrStudents);
}

void StudentData::insertData(const CsvDump &c, const qsizetype regIdx,
                             const qsizetype attemptIdx) noexcept(false) {

    // Read the registration number and attempt number from the csv-file to our
    // struct
    for (const auto &line : c.data | std::views::drop(1)) {

        bool regNrSuccess, attemptNrSuccess;
        line[regIdx].toInt(&regNrSuccess);
        const auto attemptNr = line[attemptIdx].toInt(&attemptNrSuccess);

        if (attemptNrSuccess) {

            if (!regNrSuccess) {
                // todo: show warning?
            }

            mRegistrationNr.emplaceBack(line[regIdx]);
            mAttemptNr.emplaceBack(attemptNr);
        } else {
            throw RwthOnlineParseException("Failed to parse attempt number " +
                                           line[attemptIdx].toStdString());
        }

        // Assert that the two arrays don't get out of sync
        if (mRegistrationNr.size() != mAttemptNr.size()) {
            throw RwthOnlineParseException(str::error::rwthArraySizeMismatch);
        }
    }
}
