#include "d2r_parser.hpp"

#include "csv_parser.hpp"
#include "data/csv_dump.hpp"
#include "data/data_manager.hpp"
#include "util/exception.hpp"
#include "util/strings.hpp"

#include <QList>
#include <QTextStream>

namespace impl {

bool toBool(const QString &str) {
    return QString::compare(str, "true", Qt::CaseInsensitive) == 0;
}
QString toString(const bool b) { return b ? "true" : "false"; }

} // namespace impl

struct D2rEntry {
    QString parent;
    QString key;
    QString value;
};

/**
 * \brief Calculate child depth of d2r key
 */
qsizetype getAndRemoveDepth(QString &key) {

    qsizetype depth = 0;
    while (key.startsWith("  ")) {
        ++depth;
        key = key.sliced(2);
    }
    return depth;
}

/**
 * \brief Loads a single d2r key-value-pair
 *
 * \param data Where to store the new key-value-pair
 * \param in Data source
 * \param parent Current parent, might be modified
 * \return Whether or not a new key-value-pair has been read
 */
bool loadD2rKey(QList<D2rEntry> &data, QTextStream &in, QString &parent) {
    QString line;
    in.readLineInto(&line);

    // Empty key: EOF
    if (line.isEmpty())
        return false;

    auto colonSplit = line.split(':');
    auto key = colonSplit[0];
    // Since the value could theoretically include the character :, we make sure
    // not to lose it here
    colonSplit.removeFirst();
    auto value = colonSplit.join(':');
    D2rEntry newEntry;

    const bool isChild = getAndRemoveDepth(key) > 0;
    if (isChild) {
        newEntry.parent = parent;
    } else {
        newEntry.parent = "";
        // set current key as potential parent for next key
        parent = key;
    }

    newEntry.key = key;

    // Special case for multi-line keys: they end with a "\"
    // TODO: allow \\ inside the value, and somehow keep this
    //       backwards-compatible
    if (key == "Dynexite" || key == "RWTHonline" || key == "Bonus") {

        if (value.trimmed() == "\\") {
            newEntry.value = "";
        } else {
            // Don't forget first line
            if (value != "") {
                line = value + '\n';
            } else {
                line = "";
            }

            // Read until \ or EOF
            QChar nextChar;
            in >> nextChar;
            while (nextChar != '\\' && !nextChar.isNull()) {
                line.append(nextChar);
                in >> nextChar;
            }
            newEntry.value = line;
        }
    } else {
        newEntry.value = value;
    }

    data.push_back(newEntry);

    return true;
}

QSharedPointer<ExamSave> readStringsToData(const QList<D2rEntry> &entries) {
    auto dump = QSharedPointer<ExamSave>::create();

    using namespace impl;

    // set the value for the given parent-key-combination
    for (const auto &[parent, key, value] : entries) {
        if (parent.isEmpty()) {
            if (key == "Dynexite") {
                try {
                    dump->mDynexite = DynexiteExam(*parseCsv(value));
                } catch (const std::exception &e) {
                    throw D2RParseException{
                        std::string(str::error::d2rDynexite) + e.what()};
                }
            } else if (key == "RWTHonline") {
                try {
                    dump->mStudents = StudentData(*parseCsv(value));
                } catch (const std::exception &e) {
                    throw D2RParseException{
                        std::string(str::error::d2rRwthOnline) + e.what()};
                }
            } else if (key == "Bonus") {
                try {
                    dump->mBonus = BonusPoints(*parseCsv(value));
                } catch (const std::exception &e) {
                    throw D2RParseException{std::string(str::error::d2rBonus) +
                                            e.what()};
                }
            } else if (key == "Version") {
                dump->mSettings.mVersion = value;
            }
        } else if (parent == "UserPointConfig") {
            if (key == "StepSize") {
                dump->mSettings.mUserPointConfigStepSize =
                    toStepSize(value.toDouble());
            } else if (key == "CountZerothAttempt") {
                dump->mSettings.mUserPointConfigCountZero = toBool(value);
            }
        } else if (parent == "UserGradeConfig") {
            if (key == "RequirePassAll") {
                dump->mSettings.mUserGradeConfigRequirePassAll = toBool(value);
            }
        } else if (parent == "ChangedMaxPointsClosed") {
            if (key == "ChangedMaxPoints") {
                dump->mSettings.mChangedMaxPointsClosed = value.toInt();
            } else if (key == "FollowDefault") {
                dump->mSettings.mChangedMaxPointsClosedFollowDefault =
                    toBool(value);
            }
        } else if (parent == "ChangedMaxPointsOpen") {
            if (key == "ChangedMaxPoints") {
                dump->mSettings.mChangedMaxPointsOpen = value.toInt();

            } else if (key == "FollowDefault") {
                dump->mSettings.mChangedMaxPointsOpenFollowDefault =
                    toBool(value);
            }
        } else if (parent == "UserGradeBoundariesClosed") {
            // To get the grade array index, remove the prefix "Boundary"
            QString prefix = "Boundary";
            if (key == "SetAuto") {
                dump->mSettings.mUserGradeBoundariesClosedAuto = toBool(value);
            } else if (key.contains(prefix)) {
                const auto idx = key.mid(prefix.size()).toUInt();
                dump->mSettings.mUserGradeBoundariesClosed[idx] = value.toInt();
            }
        } else if (parent == "UserGradeBoundariesOpen") {
            // To get the grade array index, remove the prefix "Boundary"
            QString prefix = "Boundary";
            if (key == "SetAuto") {
                dump->mSettings.mUserGradeBoundariesOpenAuto = toBool(value);
            } else if (key.contains(prefix)) {
                const auto idx = key.mid(prefix.size()).toUInt();
                dump->mSettings.mUserGradeBoundariesOpen[idx] = value.toInt();
            }
        }
    }

    return dump;
}

ExamSave::ExamSave(const CsvDump &dynexite, CsvDump &rwthOnline,
                   const CsvDump &bonus, const D2rSettings &settings)
    : mDynexite(dynexite), mStudents(rwthOnline), mBonus(bonus),
      mSettings(settings) {}

QSharedPointer<ExamSave> parseD2r(QString d2rString) {

    // store the resulting save on the heap, as it might get quite big
    auto save = QSharedPointer<ExamSave>::create();

    // todo: document d2r format somewhere (simplified yaml)

    QTextStream in(&d2rString);
    QString line;

    // skip ---
    {
        in.readLineInto(&line);
        if (line != "---") {
            throw D2RParseException{str::error::d2rHeader};
        }
    }

    QList<D2rEntry> readData;
    {
        QString parent;
        while (loadD2rKey(readData, in, parent)) {
        }
    }

    // Convert value strings to correct data types
    return readStringsToData(readData);
}

template <typename T>
void writeSingleSetting(QTextStream &out, const QString &name, const T &value,
                        const bool isChild = false) {
    if (isChild) {
        out << "  ";
    }
    out << name << ":" << value << '\n';
}
template <>
void writeSingleSetting<bool>(QTextStream &out, const QString &name,
                              const bool &value, const bool isChild) {
    if (isChild) {
        out << "  ";
    }
    out << name << ":" << impl::toString(value) << '\n';
}

void writeCurrentSettingsToD2r(QTextStream &out) {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    // D2R identifier
    out << "---\n";

    // Version
    writeSingleSetting(out, "Version", D2R_VERSION);

    // UserPointConfig
    writeSingleSetting(out, "UserPointConfig", "");
    {
        writeSingleSetting(
            out, "StepSize",
            as<double>(cur->mUserPointConfig.mGradeBoundaryStepSize), true);
        writeSingleSetting(
            out, "CountZerothAttempt",
            cur->mUserPointConfig.mCountZerothAttemptAsFirstAttempt, true);
    }

    // UserGradeConfig
    writeSingleSetting(out, "UserGradeConfig", "");
    {
        writeSingleSetting(out, "RequirePassAll",
                           cur->mUserGradeConfig.mRequirePassAllParts, true);
    }

    // ChangedMaxPointsClosed
    writeSingleSetting(out, "ChangedMaxPointsClosed", "");
    {
        writeSingleSetting(
            out, "ChangedMaxPoints",
            cur->mChangedMaxPointsClosed.mChangedMaxPointBoundary.mMultiple,
            true);
        writeSingleSetting(out, "FollowDefault",
                           cur->mChangedMaxPointsClosed.mFollowDefault, true);
    }

    // ChangedMaxPointsOpen
    writeSingleSetting(out, "ChangedMaxPointsOpen", "");
    {
        writeSingleSetting(
            out, "ChangedMaxPoints",
            cur->mChangedMaxPointsOpen.mChangedMaxPointBoundary.mMultiple,
            true);
        writeSingleSetting(out, "FollowDefault",
                           cur->mChangedMaxPointsOpen.mFollowDefault, true);
    }

    // UserGradeBoundaries
    for (const auto &type : {GradingType::Closed, GradingType::Open}) {
        writeSingleSetting(out, QString("UserGradeBoundaries") + toString(type),
                           "");
        {
            writeSingleSetting(
                out, "SetAuto",
                type == GradingType::Closed
                    ? cur->mClosedUserBoundaries.mSetAutomatically
                    : cur->mOpenUserBoundaries.mSetAutomatically,
                true);
            for (const auto &g : grades) {
                writeSingleSetting(
                    out, "Boundary" + QString::number(static_cast<int>(g)),
                    cur->getUserGradeBoundaries(type)[g].mMultiple, true);
            }
        }
    }

    // Dynexite CSV
    out << "Dynexite:";
    writeCsv(out, cur->mExam.mDynexite);
    out << "\\";

    // RWTHonline CSV
    out << "RWTHonline:";
    writeCsv(out, cur->mStudents.mRwthOnline);
    out << "\\";

    // Bonus CSV
    out << "Bonus:";
    writeCsv(out, cur->mBonus.mBonusPoints);
    out << "\\";
}
