#pragma once

#include "data/nodes/dynexite_exam.hpp"
#include "util/grade_boundary_step.hpp"
#include "util/points.hpp"

#include <QList>
#include <QtGlobal>

#include <cmath>
#include <ranges>

class QString;

// Since the standard library on macOS doesn't support join, we have to do
// this uglier
#define RANGE_JOIN_SUPPORTED 0

#if RANGE_JOIN_SUPPORTED
#include <utility>

/**
 * \brief Try to guess the smallest point step for this exam.
 *
 * Reads through the given ranges and returns the smallest point step found in
 * them. The actually smallest used point step might be smaller, but just never
 * happened to show up. For any found point step < 0.25, 0.25 is returned.
 *
 * \param ranges Ranges over all points.
 * \return An upper bound on the smallest point step.
 */
[[nodiscard]] constexpr GradeBoundaryStepSize
guessGradeBoundaryStepSize(std::ranges::input_range auto &&...ranges) noexcept;

/**
 * \brief Overload of \ref guessGradeBoundaryStepSize to use all relevant point
 * lists from an exam.
 */
[[nodiscard]] constexpr GradeBoundaryStepSize
guessGradeBoundaryStepSize(const DynexiteExam &exam) noexcept {
    // Search earned closed, earned open and max point lists (earned combined
    // points should be just the sum of earned closed and earned open, so no
    // need to search that as well).

    // The lists that stores these values have different types. The earned
    // points for each task are stored in `QList<QList<Points>>`. The max points
    // for all tasks are stored in a `QList<Points>`. To enable iterating over
    // all of them in one go, we flatten the `QList<QList<Points>>` into views
    // of `QList<Points>` and create a regular view for the maxPoint-`QList`.

    // It would be nice to combine the three views into one, but as of C++20
    // this is not possible yet (afaik). `views::concat` might do the trick in
    // the future.
    return guessGradeBoundaryStepSize(
        std::views::join(exam.mTaskEarnedPoints[GradingType::Closed]),
        std::views::join(exam.mTaskEarnedPoints[GradingType::Open]),
        exam.mTaskMaxPoints);
}
#else
[[nodiscard]] GradeBoundaryStepSize
guessGradeBoundaryStepSize(const DynexiteExam &exam) noexcept;
#endif

// Guess by the left-most header what file type this is
[[nodiscard]] FileType getApproximateFileType(const QString &s) noexcept;

//-----------------------------------------------------------------------
// IMPLEMENTATION
//-----------------------------------------------------------------------

namespace impl {
/**
 * \brief Returns the guessed step size for the points in the range, e.g. 1 if
 * all point-values are integers.
 */
[[nodiscard]] constexpr double
minStepSize(std::ranges::input_range auto &&range) noexcept {
    double minStep = 1.;
    for (const Points p : range) {
        if (qFuzzyIsNull(std::fmod(p.p, 1.)))
            continue;
        else if (qFuzzyIsNull(std::fmod(p.p, 0.5)))
            minStep = qMin(minStep, 0.5);
        else if (qFuzzyIsNull(std::fmod(p.p, 0.25)))
            minStep = qMin(minStep, 0.25);
        else
            minStep = qMin(minStep, std::fmod(p.p, 0.25));
    }
    return minStep;
}
} // namespace impl

#if RANGE_JOIN_SUPPORTED
[[nodiscard]] constexpr GradeBoundaryStepSize
guessGradeBoundaryStepSize(std::ranges::input_range auto &&...ranges) noexcept {

    if constexpr (sizeof...(ranges) == 0) {
        return GradeBoundaryStepSize::Min1_00;
    } else {
        const auto min = qMin(
            {impl::minStepSize(std::forward<decltype(ranges)>(ranges))...});

        // return Min0_25 for step sizes < 0.25
        if (min == 1.0)
            return GradeBoundaryStepSize::Min1_00;
        else if (min == 0.5)
            return GradeBoundaryStepSize::Min0_50;
        else
            return GradeBoundaryStepSize::Min0_25;
    }
}
#else
[[nodiscard]] GradeBoundaryStepSize
guessGradeBoundaryStepSize(const QList<Points> &allPointValues) noexcept;
#endif
