#include "load_file_view.hpp"

#include "util/storage_helper.hpp"
#include "util/strings.hpp"

#include <QFileDialog>
#include <QFrame>
#include <QMessageBox>
#include <QSpacerItem>
#include <QString>
#include <QStyle>

#include <array>
#include <cassert>

#define HOME_MACHINE 1

QString getFileNameDesc(const FileType t) {
    switch (t) {
    case FileType::dynexite:
        return "Dynexite";
    case FileType::rwth:
        return "RWTHonline";
    case FileType::d2r:
        return "D2R";
    case FileType::bonus:
        return "Bonus points (optional)";
    default:
        assert(false);
        return "Unknown";
    }
}
[[nodiscard]] int getFileHelp(const FileType t, QGridLayout &layout,
                              const QStyle *style, QWidget *parent) {

    // Question mark icon
    const auto iconLabel = new QLabel("", parent);
    iconLabel->setPixmap(
        style->standardIcon(QStyle::SP_MessageBoxInformation).pixmap(16));
    layout.addWidget(iconLabel, 0, 0);

    if (static_cast<std::size_t>(t) >= str::ui::input::helpTexts.size()) {
        assert(false);
        return 1;
    }

    layout.addWidget(
        new QLabel(str::ui::input::helpTexts[static_cast<std::size_t>(t)],
                   parent),
        0, 1);

    // Separator line
    const auto line = new QFrame(parent);
    line->setFrameShape(QFrame::HLine);
    layout.addItem(
        new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Minimum), 0,
        2);

    layout.addWidget(line, 1, 0, 1, 3);
    return 2;
}

LoadFileView::LoadFileView(const FileType type, QWidget *parent)
    : QGroupBox(getFileNameDesc(type), parent), mInputModel(type) {

    // todo: set this, but it also does some annoying stuff - don't do them.
    // mFileNameLabel->setWordWrap(true);

    connect(&mBrowseButton, &QPushButton::clicked, this,
            &LoadFileView::handleBrowseButton);

    // First, an explanatory text
    auto nextRow = getFileHelp(type, mOuterLayout, style(), this);

    // The file selection is surrounded by two spacers, so that it is in the
    // middle of the remaining space
    mOuterLayout.addItem(
        new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Expanding),
        nextRow++, 0, 1, 3);

    // the elements of each file selection are arranged horizontally within one
    // row of the outer layout
    mOuterLayout.addLayout(&mHLayout, nextRow++, 0, 1, 3);
    mHLayout.addWidget(&mFileNameLabel);
    mHLayout.addWidget(&mFileParseStatusIcon);
    mHLayout.addWidget(&mBrowseButton);

    mOuterLayout.addItem(
        new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Expanding),
        nextRow++, 0, 1, 3);
    setLayout(&mHLayout);
}

void LoadFileView::handleBrowseButton() {
    // window captions for the file selection dialog for each file type
    static constexpr EnumStorage<const char *, FileType> captionTexts = {
        "Open exam results file from dynexite",
        "Open student info file from RWTHonline",
        "Open eTest bonus points file", "Open all-in-one (d2r) file"};

    // let the user select a file
    const auto fileName = QFileDialog::getOpenFileName(
        this, captionTexts[mInputModel.mType], "",
        mInputModel.mType == FileType::d2r ? str::ui::input::d2rFilter
                                           : str::ui::input::csvFilter);

    // if the user didn't select any files, don't change anything
    if (!fileName.isEmpty()) {
        // write the path to the selected file to the label of the corresponding
        // file type
        mFileNameLabel.setText(fileName);

        const auto error = mInputModel.loadCsv(mFileNameLabel.text());
        if (error.mEc == FileOpenErrorCode::Success) {

            // Display warnings if there are any
            if (!error.mWarnings.empty()) {
                QString errorText = str::error::readingFileWarning;
                for (const auto &s : error.mWarnings) {
                    errorText += '\n' + s;
                }
                QMessageBox::warning(this, str::title::readingFileWarning,
                                     errorText);
            }

            setFileStatus(FileSelectState::ok);
        } else {

            QMessageBox::warning(this, str::title::readingFileError,
                                 error.mDetails);

            setFileStatus(FileSelectState::error);
        }
    }
}

void LoadFileView::reset() {
    // remove input from the labels displaying the selected files, reset icons,
    // unload csv-file
    mFileNameLabel.setText("");

    // We deliberately keep the current data loaded in the model, so that we can
    // reset the view and extract the data from the model (which clears it
    // there)

    setFileStatus(FileSelectState::empty);
}

void LoadFileView::tempSetDeveloperPaths() {
#if defined(DYNEXITE_EXAMPLE_FILE_PATH) && defined(RWTHONLINE_EXAMPLE_FILE_PATH)
    if (mInputModel.mType == FileType::dynexite) {
        mFileNameLabel.setText(DYNEXITE_EXAMPLE_FILE_PATH);
    } else {
        mFileNameLabel.setText(RWTHONLINE_EXAMPLE_FILE_PATH);
    }

    if (mInputModel.loadCsv(mFileNameLabel.text()).mEc ==
        FileOpenErrorCode::Success) {
        setFileStatus(FileSelectState::ok);
    } else {
        setFileStatus(FileSelectState::error);
    }
#endif
}

void LoadFileView::setFileStatus(const FileSelectState status) {
    // icons corresponding to each file select state
    static constexpr EnumStorage<QStyle::StandardPixmap, FileSelectState>
        icons = {
            QStyle::SP_MessageBoxQuestion, // empty, no icon, not used
            QStyle::SP_MessageBoxCritical,
            QStyle::SP_DialogOkButton // todo: find better icon
        };

    // update status
    mFileState = status;

    // update icon
    if (status != FileSelectState::empty) {
        mFileParseStatusIcon.setPixmap(
            style()->standardIcon(icons[status]).pixmap(16));
    } else {
        mFileParseStatusIcon.setText(" ");
    }
    emit modelChange();
}
