/**
 * \file main_window.cpp
 * \brief Definition of the \ref MainWindow member functions.
 */

#include "main_window.hpp"

#include "export_rwthonline.hpp"
#include "grade_tab.hpp"
#include "students_tab.hpp"
#include "task_tab.hpp"

#include "data/data_manager.hpp"
#include "file_io/d2r_parser.hpp"
#include "file_io/file.hpp"
#include "util/defaults.hpp"
#include "util/enums.hpp"
#include "util/font.hpp"
#include "util/grade.hpp"
#include "util/grade_boundary.hpp"
#include "util/grade_boundary_step.hpp"

#include <QAction>
#include <QApplication>
#include <QClipboard>
#include <QDesktopServices>
#include <QFileDialog>
#include <QLabel>
#include <QMenuBar>
#include <QString>
#include <QUrl>

#include <cmath>

// Surround the given string evenly with spaces, until it has a size of `size`
QString padToSize(QString str, const int size) {
    const auto spacesLeft = static_cast<int>(
        qMax(0., std::floor(static_cast<double>(size - str.size()) / 2.)));
    const auto spacesRight = static_cast<int>(
        qMax(0., std::ceil(static_cast<double>(size - str.size()) / 2.)));

    str.prepend(QString(spacesLeft, ' '));
    str.append(QString(spacesRight, ' '));

    return str;
}

void printBoundaries(QString &str, const GradingType type, const bool hasType,
                     const GradeBoundaryStepSize stepSize,
                     const int decimalPlaces) {

    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    if (hasType) {
        str += QString(toString(type)) +
               " part:\n"
               "Grade --- Min. points --- Max. points\n";

        auto lastGradeBoundary =
            cur->mChangedMaxPointsOpen.mChangedMaxPointBoundary;

        for (const auto g : grades) {

            const auto minBoundary =
                toPoints(cur->getUserGradeBoundaries(type)[g], stepSize).p;

            // For max points of 1.0, we use exam max points, not changed max
            // points.
            // For other max points, we go one grade boundary step below the
            // next higher grade.
            const auto maxBoundary =
                g == Grade::grade1_0 ? cur->mExam.mPointsMax[type].p
                                     : toPoints(lastGradeBoundary, stepSize).p -
                                           as<double>(stepSize);

            str += padToSize(toString(g), 7) + ":" +
                   padToSize(QString::number(minBoundary, 'f', decimalPlaces),
                             15) +
                   "-" +
                   padToSize(QString::number(maxBoundary, 'f', decimalPlaces),
                             13) +
                   "\n";
            lastGradeBoundary = cur->getUserGradeBoundaries(type)[g];
        }

        str += "\n";
    }
}

void copyGradingSchemeToClipboard() {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    auto *clipboard = QGuiApplication::clipboard();

    const bool hasClosed =
        cur->mExam.mPointsMax[GradingType::Closed] > Points{0.};
    const bool hasOpen = cur->mExam.mPointsMax[GradingType::Open] > Points{0.};

    QString gradingSchemeText = "";

    const auto stepSize = cur->mUserPointConfig.mGradeBoundaryStepSize;
    const auto decimalPlaces = nrDecimalPlaces(stepSize);

    printBoundaries(gradingSchemeText, GradingType::Open, hasOpen, stepSize,
                    decimalPlaces);
    printBoundaries(gradingSchemeText, GradingType::Closed, hasClosed, stepSize,
                    decimalPlaces);

    if (!hasOpen && !hasClosed) {
        gradingSchemeText +=
            "Selected exam has neither open nor closed points.";
    }

    clipboard->setText(gradingSchemeText);
}

void exportD2R(QWidget *parent) {
    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    const auto exportLocationString = QFileDialog::getSaveFileName(
        parent, "Save data all-in-one (d2r)", "",
        "All-in-one (D2R) file (*.d2r);;All files (*.*)");

    if (!exportLocationString.isEmpty()) {
        auto [file, stream] = openFileToWrite(exportLocationString);
        writeCurrentSettingsToD2r(*stream);
    }
}

MainWindow::MainWindow(QWidget *parent)
    : QMainWindow(parent), mInputWindow(*this, parent), mBoundaryWindow(parent),
      mViewSettings(parent), mStatisticsWindow(parent) /*,
       mGradeTab(mViewSettings), mTaskTab(), mStudentsTab()*/
{

    setWindowTitle("DynAMITE");

    setCentralWidget(&mTabs);
    // handleCommandLineArgs();

    mGradeTab = new GradeTab{mViewSettings};
    mTaskTab = new TaskTab{};
    mStudentsTab = new StudentsTab{};
    mTabs.addTab(mGradeTab, "Grading");
    mTabs.addTab(mTaskTab, "Tasks");
    mTabs.addTab(mStudentsTab, "Students");

    // Create handle for all windows for operations that need to iterate over
    // all of them
    mSubWindowPtrs[0] = &mInputWindow;
    mSubWindowPtrs[1] = &mBoundaryWindow;
    mSubWindowPtrs[2] = &mViewSettings;
    mSubWindowPtrs[3] = &mStatisticsWindow;

    // Set appropriate flags for all small openable windows: not maximizable,
    // stay on top, closable
    for (const auto &window : mSubWindowPtrs) {
        window->setWindowFlags(
            Qt::WindowStaysOnTopHint | Qt::WindowCloseButtonHint |
            Qt::WindowMinimizeButtonHint | Qt::MSWindowsFixedSizeDialogHint);
        window->setWindowFlags(mBoundaryWindow.windowFlags() &
                               ~Qt::WindowMaximizeButtonHint);
    }

    // start input file selection
    mInputWindow.show();

    createMenuBar();

    connect(&mTabs, &QTabWidget::currentChanged, this, &MainWindow::tabChanged);
}

void MainWindow::resetToInputFiles(const DynexiteExam &exam,
                                   const StudentData &students,
                                   const BonusPoints &bonus,
                                   const D2rSettings &settings) {

    UserPointConfig pc;
    pc.mGradeBoundaryStepSize = settings.mUserPointConfigStepSize;
    if (DataManager::loadNewExam(exam, students, bonus, UserGradeConfig{},
                                 pc)) {
        DataManager::makeNewCurrent();

        DataManager::applySettings(settings);

        // close input window and show main window
        mInputWindow.hide();
        showMaximized();

        updateMenuItems();

        // Change font size to adapt to new bar sizes
        handleResize();
    }
}

void MainWindow::menuNewFile() { mInputWindow.show(); }

void MainWindow::menuOpenBoundaries() { mBoundaryWindow.show(); }

void MainWindow::menuOpenViewSettings() { mViewSettings.show(); }

void MainWindow::menuOpenStatistics() { mStatisticsWindow.show(); }

void MainWindow::tabChanged(int) {
    // handle resize here again, because sizes of non-displayed tabs are not
    // calculated correctly
    handleResize();
}

void MainWindow::closeEvent(QCloseEvent *event) {
    for (const auto &window : mSubWindowPtrs) {
        window->hide();
    }

    QMainWindow::closeEvent(event);
}

void MainWindow::resizeEvent(QResizeEvent *event) {
    QMainWindow::resizeEvent(event);

    handleResize();

    // qDebug() << "Is maximized? " <<
    // windowState().testFlag(Qt::WindowMaximized);

    // auto oldSize = this->size();
    // QTimer::singleShot(1000, this,
    //                    [this, oldSize]() { this->resize(oldSize); });

    // static bool wasMaximized = windowState().testFlag(Qt::WindowMaximized);
    // const bool isMaximized = windowState().testFlag(Qt::WindowMaximized);

    // if (wasMaximized && !isMaximized) {
    //     // update flag before call to resize
    //     wasMaximized = isMaximized;

    //    auto oldSize = this->size();
    //    oldSize.setWidth(oldSize.width() - 1);
    //    this->resize(oldSize);
    //}
    // wasMaximized = isMaximized;

    // If the flag "WindowMaximized" is set here, that means the window WAS
    // maximized before the current resize, so it isn't anymore
    // if (windowState().testFlag(Qt::WindowMaximized)) {

    //    // todo: explain this madness
    //    auto mySize = this->size();
    //    mySize.setWidth(mySize.width() - 1);
    //    QTimer::singleShot(0, this, [this, mySize]() { this->resize(mySize);
    //    });
    //}

    // static bool doubleExecution = true;
    // if (doubleExecution) {
    //     auto oldSize = this->size();
    //     if (oldSize.width() < 2300) {

    //        doubleExecution = false;
    //        oldSize.setWidth(oldSize.width() - 1);
    //        QTimer::singleShot(0, this,
    //                           [this, oldSize]() { this->resize(oldSize); });
    //    }

    //} else {
    //    doubleExecution = true;
    //}
}

void MainWindow::handleResize() {

    static FontSize defaultFontSize;

    static bool firstExecution = true;
    if (firstExecution) {

        const auto defaultFont = QLabel{}.font();

        // Default font size might be given in points or in pixels
        if (defaultFont.pixelSize() > 0) {
            defaultFontSize = {static_cast<qreal>(defaultFont.pixelSize()),
                               FontSizeUnit::Pixel};
        } else {
            defaultFontSize = {defaultFont.pointSizeF(), FontSizeUnit::Point};
        }

        firstExecution = false;
    }

    // Find out new font size (not smaller than default font)
    const auto possibleLargeFontSize = mGradeTab->getFontPxSize(this->size());

    // Large font size may not be smaller than the default font size
    const auto largeFontSize = qMax(possibleLargeFontSize, defaultFontSize);

    // Small font size is a fixed ratio smaller than the large font size, but
    // also not smaller than defaultFontSize
    const auto smallFontSize =
        qMax(largeFontSize * defaults::smallFontSizeRatio, defaultFontSize);

    // Set font size of all labels, text boxes etc
    auto defaultFont = QApplication::font();
    setFontSize(defaultFont, smallFontSize);
    QApplication::setFont(defaultFont);

    // Set font size for legends and adapt axis spacing
    mGradeTab->handleResize(largeFontSize, smallFontSize);

    mTaskTab->handleResize(this->size(), largeFontSize);
}

void MainWindow::updateMenuItems() const {
    const auto nrUnknownStudents = getNrUnknownStudents();

    mExportUnknownStudentsAct->setText("Manual: Unknown Students (" +
                                       QString::number(nrUnknownStudents) +
                                       ")");
    mExportUnknownStudentsAct->setEnabled(nrUnknownStudents > 0);
}

void MainWindow::createMenuBar() {

    /***************** Exam **************/
    mFileMenu = menuBar()->addMenu("&Exam");

    // "New": loads a new exam
    {
        const auto newAct = new QAction("&Open", this);
        newAct->setShortcuts(QKeySequence::Open);
        connect(newAct, &QAction::triggered, this, &MainWindow::menuNewFile);
        mFileMenu->addAction(newAct);
    }

    // "Options": change the required points for e.g. 2.7, change requirePassAll
    {
        const auto boundaryAct = new QAction("Options", this);
        boundaryAct->setShortcuts(QKeySequence::Preferences);
        connect(boundaryAct, &QAction::triggered, this,
                &MainWindow::menuOpenBoundaries);
        mFileMenu->addAction(boundaryAct);
    }

    // "Statistics": show more detailed statistics like number of attended
    // students
    {
        const auto statisticsAct = new QAction("Statistics", this);
        connect(statisticsAct, &QAction::triggered, this,
                &MainWindow::menuOpenStatistics);
        mFileMenu->addAction(statisticsAct);
    }

    // "Tasks": show more detailed task info (temporarily here)
    //{
    //    const auto tasksAct = new QAction("Tasks", this);
    //    connect(tasksAct, &QAction::triggered, this,
    //            &MainWindow::menuOpenTasks);
    //    mFileMenu->addAction(tasksAct);
    //}

    /***************** View **************/
    mViewMenu = menuBar()->addMenu("&View");

    // "View settings": change visual options
    {
        const auto viewSettingsAct = new QAction("Settings", this);
        connect(viewSettingsAct, &QAction::triggered, this,
                &MainWindow::menuOpenViewSettings);
        mViewMenu->addAction(viewSettingsAct);
    }

    /**************** Export *************/
    mExportMenu = menuBar()->addMenu("E&xport");

    // Export grading scheme to clipboard
    {
        mExportGradingMenu = mExportMenu->addMenu("Grading scheme");
        const auto exportGradingSchemeClipboardAct =
            new QAction("To Clipboard", this);
        connect(exportGradingSchemeClipboardAct, &QAction::triggered,
                copyGradingSchemeToClipboard);
        mExportGradingMenu->addAction(exportGradingSchemeClipboardAct);
    }

    // Export students
    {
        mExportStudentsMenu = mExportMenu->addMenu("Results");
        // Export known students as CSV
        {
            const auto exportKnownStudentsAct =
                new QAction("RWTHonline: Known Students", this);
            connect(exportKnownStudentsAct, &QAction::triggered,
                    [this] { exportKnownStudents(this); });
            mExportStudentsMenu->addAction(exportKnownStudentsAct);
        }

        // Export unknown students as CSV
        {
            mExportUnknownStudentsAct =
                new QAction("Manual: Unknown Students", this);
            connect(mExportUnknownStudentsAct, &QAction::triggered,
                    [this] { exportUnknownStudents(this); });
            mExportStudentsMenu->addAction(mExportUnknownStudentsAct);
        }
    }

    // Export to d2r
    {
        const auto exportD2RAct = new QAction("D2R", this);
        connect(exportD2RAct, &QAction::triggered, [this] { exportD2R(this); });
        mExportMenu->addAction(exportD2RAct);
    }

    /**************** Help *************/
    mHelpMenu = menuBar()->addMenu("&Help");

    // Open Handreichung / Guidelines
    {
        mOpenGuidelinesMenu = mHelpMenu->addMenu("Open grading guidelines");

        // German
        {
            const auto mOpenGuidelinesGerman = new QAction("German", this);
            connect(mOpenGuidelinesGerman, &QAction::triggered, [] {
                QDesktopServices::openUrl(
                    QUrl("https://www.rwth-aachen.de/global/"
                         "show_document.asp?id=aaaaaaaaacxybai"));
            });
            mOpenGuidelinesMenu->addAction(mOpenGuidelinesGerman);
        }

        // English
        {
            const auto mOpenGuidelinesEnglish = new QAction("English", this);
            connect(mOpenGuidelinesEnglish, &QAction::triggered, [] {
                QDesktopServices::openUrl(
                    QUrl("https://www.rwth-aachen.de/global/"
                         "show_document.asp?id=aaaaaaaaaejylxi"));
            });
            mOpenGuidelinesMenu->addAction(mOpenGuidelinesEnglish);
        }
    }

    // "Version": show version (directly in the menu item, not in a popup
    // window)
    {
        const auto versionAct =
            new QAction(QString("Version: ") + D2R_VERSION, this);
        connect(versionAct, &QAction::triggered, this, [] {});
        versionAct->setDisabled(true);
        mHelpMenu->addAction(versionAct);
    }
}

/* void MainWindow::handleCommandLineArgs() const {
    // get list of command line arguments
    QStringList fileNames = QCoreApplication::arguments();

    // skip first argument (path of executable)
    if (fileNames.size() > 1) {
        mD2RFileChooserTab->newInputFile(fileNames[1]);
    }
}*/
