#include "min_max_point_screws.hpp"

#include "data/data_manager.hpp"
#include "util/font.hpp"
#include "util/grade_boundary.hpp"
#include "util/grade_boundary_step.hpp"
#include "util/strings.hpp"
#include "view/gui/custom_slider.hpp"

#include <QDoubleSpinBox>

#include <array>
#include <cassert>
#include <cstddef>

MinMaxPointScrewColumn::MinMaxPointScrewColumn(const GradingType grading,
                                               QWidget *parent)
    : QVBoxLayout(parent), mParent(parent), mGrading(grading) {

    assert(grading == GradingType::Closed || grading == GradingType::Open);

    const QString labelExt =
        grading == GradingType::Closed ? " (Closed)" : " (Open)";

    createLabel(grading, 0, "Maximal Points" + labelExt, 0, false, mParent);
    createSlider(grading, 0, 1, mParent);

    addSpacing(20);

    createLabel(grading, 1, "Minimal Points" + labelExt, 2, true, mParent);
    createSlider(grading, 1, 3, mParent);

    // receive update when points change
    connect(&DataManager::getInstance(), &DataManager::gradesChanged, this,
            &MinMaxPointScrewColumn::gradesChanged);
}

void MinMaxPointScrewColumn::setWidgetsEnabled(const bool enable) const {
    for (auto i = 0u; i < std::size(mSpinBoxes); ++i) {
        mSpinBoxes[i]->setEnabled(enable);
    }
    for (auto i = 0u; i < std::size(mSliders); ++i) {
        mSliders[i]->setEnabled(enable);
    }
}

void MinMaxPointScrewColumn::createLabel(const GradingType grading,
                                         const int labelIdx,
                                         const QString &text, const int row,
                                         const bool isMin, QWidget *parent) {
    mHLayouts[row] = new QHBoxLayout(parent);

    mDescriptions[labelIdx] = new QLabel(text, parent);
    mSpinBoxes[row] = new QDoubleSpinBox(parent);

    mHLayouts[row]->addWidget(mDescriptions[labelIdx]);
    createQuestionIcon(grading, row, isMin, parent);
    mHLayouts[row]->addWidget(mSpinBoxes[row]);

    connect(mSpinBoxes[row],
            QOverload<double>::of(&QDoubleSpinBox::valueChanged), this,
            [this, row, grading](const double value) {
                spinboxChange(grading, row, value);
            });

    addLayout(mHLayouts[row]);
}

void MinMaxPointScrewColumn::createSlider(const GradingType grading,
                                          const int sliderIdx, const int row,
                                          QWidget *parent) {

    mHLayouts[row] = new QHBoxLayout(parent);

    mSliders[sliderIdx] = new CustomSlider(parent);
    mSpinBoxes[row] = new QDoubleSpinBox(parent);
    mSpinBoxes[row]->setSuffix("%");
    mSpinBoxes[row]->setDecimals(2); // todo: make configurable
    mSpinBoxes[row]->setKeyboardTracking(false);

    mHLayouts[row]->addWidget(mSliders[sliderIdx]);
    mHLayouts[row]->addWidget(mSpinBoxes[row]);

    connect(mSliders[sliderIdx], &CustomSlider::valueChanged, this,
            [this, sliderIdx, grading](const int value) {
                sliderChange(grading, sliderIdx, value);
            });
    if (grading == GradingType::Closed) {
        connect(mSpinBoxes[row],
                QOverload<double>::of(&QDoubleSpinBox::valueChanged), this,
                [this, row](const double value) {
                    spinboxChange(GradingType::Closed, row, value);
                });
    } else {
        connect(mSpinBoxes[row],
                QOverload<double>::of(&QDoubleSpinBox::valueChanged), this,
                [this, row](const double value) {
                    spinboxChange(GradingType::Open, row, value);
                });
    }

    addLayout(mHLayouts[row]);
}

void MinMaxPointScrewColumn::createQuestionIcon(const GradingType grading,
                                                const int row, const bool isMin,
                                                QWidget *parent) {
    // Only add the question icon to the closed column and the min row
    if (grading == GradingType::Closed && isMin) {
        // show info both as tooltip and when clicked on
        mQuestionIcon = new HelpWidget(parent, "Closed passing mark", "");

        mHLayouts[row]->addWidget(mQuestionIcon);
    } else {
        // Leave the same amount of space
        const auto uselessButton = new QPushButton(parent);
        uselessButton->setFlat(true);
        uselessButton->setEnabled(false);
        mHLayouts[row]->addWidget(uselessButton);
    }
}

void MinMaxPointScrewColumn::setSpinboxRange(
    const int row, const GradeBoundary min, const GradeBoundary max,
    const GradeBoundary start, const GradeBoundaryStepSize minPointStep) const {

    // don't trigger the valueChange-callback
    QSignalBlocker blocker(mSpinBoxes[row]);

    mSpinBoxes[row]->setRange(toPoints(min, minPointStep).p,
                              toPoints(max, minPointStep).p);
    mSpinBoxes[row]->setValue(toPoints(start, minPointStep).p);

    mSpinBoxes[row]->setDecimals(nrDecimalPlaces(minPointStep));
    mSpinBoxes[row]->setSingleStep(as<double>(minPointStep));
}

void MinMaxPointScrewColumn::setSliderRange(
    const int row, const int sliderIdx, const GradeBoundary rangeMin,
    const GradeBoundary rangeMax, const GradeBoundary start,
    const GradeBoundary maxPoints) const {

    // don't trigger the valueChange-callback
    QSignalBlocker blocker1(mSliders[sliderIdx]);
    QSignalBlocker blocker2(mSpinBoxes[row]);

    /* todo: default slider only slides over ints, doesn't quite work if grade
     * boundary step is no int */
    /* todo: slide over all possible points, not just all possible integer
     * points */
    mSliders[sliderIdx]->setRange(rangeMin.mMultiple, rangeMax.mMultiple);
    mSliders[sliderIdx]->setValue(start.mMultiple);

    mSpinBoxes[row]->setRange(rangeMin / maxPoints * 100.,
                              rangeMax / maxPoints * 100.);
    mSpinBoxes[row]->setValue(start / maxPoints * 100.);
    const auto nrViablePositions = rangeMax.mMultiple - rangeMin.mMultiple;
    const auto singleStepSize =
        (mSpinBoxes[row]->maximum() - mSpinBoxes[row]->minimum()) /
        static_cast<double>(nrViablePositions);
    mSpinBoxes[row]->setSingleStep(singleStepSize);
}

void MinMaxPointScrewColumn::gradesChanged() {

    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    const auto totalPoints =
        mGrading == GradingType::Closed
            ? cur->mChangedMaxPointsClosed.mOriginalMaxPointBoundary
            : cur->mChangedMaxPointsOpen.mOriginalMaxPointBoundary;
    const auto boundary4_0 = mGrading == GradingType::Closed
                                 ? cur->mClosedPassingLimit.mMaxPassingLimit
                                 : cur->mOpenPassingLimit.mMaxPassingLimit;

    // todo: are these necessary in this calculation?
    // todo: update in grade_update the range of the spinbox/slider
    const auto currentTotalPoints =
        mGrading == GradingType::Closed
            ? cur->mChangedMaxPointsClosed.mChangedMaxPointBoundary
            : cur->mChangedMaxPointsOpen.mChangedMaxPointBoundary;
    const auto currentBoundary4_0 =
        mGrading == GradingType::Closed
            ? cur->mClosedUserBoundaries.mBoundaries[Grade::grade4_0]
            : cur->mOpenUserBoundaries.mBoundaries[Grade::grade4_0];

    const auto minPointStep = cur->mUserPointConfig.mGradeBoundaryStepSize;

    // Spinbox
    setSpinboxRange(0, boundary4_0, totalPoints, currentTotalPoints,
                    minPointStep);
    setSpinboxRange(2, GradeBoundary{0}, boundary4_0, currentBoundary4_0,
                    minPointStep);

    // Slider
    /* todo: maybe position slider next to points if it controls points, or next
     * to percent if it controls percent */
    setSliderRange(1, 0, boundary4_0, totalPoints, currentTotalPoints,
                   totalPoints);
    setSliderRange(3, 1, GradeBoundary{0}, boundary4_0, currentBoundary4_0,
                   currentTotalPoints);

    // Question icon hover text
    if (mGrading == GradingType::Closed) {
        QString questionIconHoverText;
        // Use rich text here for smart line breaks
        if (mSpinBoxes[3]->maximum() > 59) {
            questionIconHoverText = QString(str::control::richTextStart) +
                                    str::info::sixtyPercentHover1 +
                                    str::control::richTextEnd;
        } else {
            questionIconHoverText =
                QString(str::control::richTextStart) +
                str::info::sixtyPercentHover1 + "<br><br>" +
                str::info::sixtyPercentHover2 +
                QString::number(mSpinBoxes[3]->maximum(), 'f', 2) + "%." +
                str::control::richTextEnd;
        }
        mQuestionIcon->setText(questionIconHoverText);
    }
}

void MinMaxPointScrewColumn::sliderChange(const GradingType grading,
                                          const int sliderIdx,
                                          const int value) {

    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    // update spinbox
    const auto spinboxIdx = sliderIdx == 0 ? 0 : 2;

    const GradeBoundary realValue{value};
    const auto stepSize = cur->mUserPointConfig.mGradeBoundaryStepSize;
    const auto max = grading == GradingType::Closed
                         ? cur->mChangedMaxPointsClosed.mChangedMaxPointBoundary
                         : cur->mChangedMaxPointsOpen.mChangedMaxPointBoundary;

    {
        QSignalBlocker blocker1(mSpinBoxes[spinboxIdx]);
        QSignalBlocker blocker2(mSpinBoxes[spinboxIdx + 1]);

        mSpinBoxes[spinboxIdx]->setValue(toPoints(realValue, stepSize).p);
        mSpinBoxes[spinboxIdx + 1]->setValue(toPoints(realValue, stepSize) /
                                             toPoints(max, stepSize) * 100.);
    }

    if (sliderIdx == 0) {
        DataManager::maxChange(grading, realValue);
    } else {
        DataManager::boundaryChange(grading, Grade::grade4_0, realValue);
    }
}

void MinMaxPointScrewColumn::spinboxChange(GradingType grading, int spinboxIdx,
                                           double value) {
    const bool isPercent = spinboxIdx % 2;
    const auto sliderIdx = spinboxIdx <= 1 ? 0 : 1;

    const auto cur = DataManager::getCurrent();
    if (!cur)
        return;

    const auto stepSize = cur->mUserPointConfig.mGradeBoundaryStepSize;
    const auto rounding = cur->mUserPointConfig.mGradeBoundaryRounding;
    const auto max = toPoints(
        [isMax = sliderIdx == 0, grading, cur] {
            if (isMax) {
                return grading == GradingType::Closed
                           ? cur->mChangedMaxPointsClosed
                                 .mOriginalMaxPointBoundary
                           : cur->mChangedMaxPointsOpen
                                 .mOriginalMaxPointBoundary;
            } else {
                return grading == GradingType::Closed
                           ? cur->mChangedMaxPointsClosed
                                 .mChangedMaxPointBoundary
                           : cur->mChangedMaxPointsOpen
                                 .mChangedMaxPointBoundary;
            }
        }(),
        stepSize);

    GradeBoundary boundary;
    if (isPercent) {
        boundary = GradeBoundary{max * (value / 100.), stepSize,
                                 GradeBoundaryRounding::Nearest};

        const auto pointSpinboxIdx = static_cast<std::size_t>(spinboxIdx) - 1;

        {
            QSignalBlocker blocker1(mSpinBoxes[pointSpinboxIdx]);
            QSignalBlocker blocker2(mSliders[sliderIdx]);
            // QSignalBlocker blocker3(mSpinBoxes[2 * sliderIdx + 1]);

            mSpinBoxes[pointSpinboxIdx]->setValue(
                toPoints(boundary, stepSize).p);
            mSliders[sliderIdx]->setValue(boundary.mMultiple);
            // mSpinBoxes[2 * sliderIdx + 1]->setValue(value);
        }

    } else {
        boundary = GradeBoundary{Points{value}, stepSize, rounding};

        const auto percent = value / max.p * 100.;

        const auto percentSpinboxIdx = static_cast<std::size_t>(spinboxIdx) + 1;

        {
            QSignalBlocker blocker1(mSpinBoxes[percentSpinboxIdx]);
            QSignalBlocker blocker2(mSliders[sliderIdx]);
            // QSignalBlocker blocker3(mSpinBoxes[2 * sliderIdx]);

            mSpinBoxes[percentSpinboxIdx]->setValue(percent);
            mSliders[sliderIdx]->setValue(boundary.mMultiple);
            // mSpinBoxes[2 * sliderIdx]->setValue(value);
        }
    }

    if (sliderIdx == 0) {
        DataManager::maxChange(grading, boundary);
    } else {
        DataManager::boundaryChange(grading, Grade::grade4_0, boundary);
    }
}

inline constexpr int avgRow = 0;
inline constexpr int avgPassedRow = 1;
inline constexpr int nrPassedRow = 2;

inline constexpr int textColumn = 0;
inline constexpr int avgNumberColumn = 1;

MinMaxPointScrews::MinMaxPointScrews(QWidget *parent)
    : QWidget(parent), mAvgLabel("Average grade (Attended):"),
      mAvgPassedLabel("Average grade (Passed):"),
      mNrPassedLabel("Attended students passed:") {

    // Even space above and below the screw columns, avg at the very bottom
    mVLayout.addItem(
        new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Expanding));

    createScrewColumns();

    mVLayout.addItem(
        new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Expanding));

    createAvgLabels();

    // receive update when grades change
    connect(&DataManager::getInstance(), &DataManager::gradesChanged, this,
            &MinMaxPointScrews::gradesChanged);
}

void MinMaxPointScrews::setEnabled(const GradingType type,
                                   const bool enabled) const {
    if (type == GradingType::Closed) {
        mClosedColumn->setWidgetsEnabled(enabled);
    } else {
        mOpenColumn->setWidgetsEnabled(enabled);
    }
}

void MinMaxPointScrews::gradesChanged() {

    const auto cur = DataManager::getCurrent();
    if (!cur) {
        mAverageGrade.setText("-");
        mAverageGradePassed.setText("-");
        mNrPassed.setText("- of -");
        return;
    }

    // Number of decimal places for average grade is always 2, independent of
    // the grade boundary step size
    mAverageGrade.setText(
        "" + QString::number(cur->mGradeAverage.mAverageGrade, 'f', 2));
    mAverageGradePassed.setText(
        "" + QString::number(cur->mGradeAverage.mAverageGradePassed, 'f', 2));
    {
        const auto &nrPassed = cur->mGradeAverage.mNrPassed;
        const auto &nrAttended = cur->mGrades.mGrades.size();
        if (nrAttended == 0) {
            mNrPassed.setText("- of -");
        } else {

            mNrPassed.setText(
                "" + QString::number(nrPassed) + " of " +
                QString::number(nrAttended) + " (" +
                QString::number(static_cast<double>(nrPassed) /
                                    static_cast<double>(nrAttended) * 100.,
                                'f', 2) +
                "%)");
        }
    }
}

void MinMaxPointScrews::createScrewColumns() {
    mVLayout.addLayout(&mHLayout);

    mHLayout.addSpacing(10);
    mHLayout.addStretch(1);

    // todo: parent
    mOpenColumn = new MinMaxPointScrewColumn(GradingType::Open);
    mHLayout.addLayout(mOpenColumn, 5);
    mHLayout.addSpacing(20);
    mHLayout.addStretch(2);

    // todo: parent
    mClosedColumn = new MinMaxPointScrewColumn(GradingType::Closed);
    mHLayout.addLayout(mClosedColumn, 5);
    mHLayout.addSpacing(10);
    mHLayout.addStretch(1);
}

void MinMaxPointScrews::changeFontSize(const FontSize newFontSize) {

    auto font = mAverageGrade.font();
    setFontSize(font, newFontSize);

    for (const auto &label : {&mAverageGrade, &mAverageGradePassed, &mNrPassed,
                              &mAvgLabel, &mAvgPassedLabel, &mNrPassedLabel}) {
        label->setFont(font);
    }
}

void MinMaxPointScrews::createAvgLabels() {
    mVLayout.addLayout(&mAvgLayout);

    // average text labels
    mAvgLayout.addWidget(&mAvgLabel, avgRow, textColumn);
    mAvgLayout.addWidget(&mAvgPassedLabel, avgPassedRow, textColumn);
    mAvgLayout.addWidget(&mNrPassedLabel, nrPassedRow, textColumn);

    // average value labels
    mAvgLayout.addWidget(&mAverageGrade, avgRow, avgNumberColumn);
    mAvgLayout.addWidget(&mAverageGradePassed, avgPassedRow, avgNumberColumn);
    mAvgLayout.addWidget(&mNrPassed, nrPassedRow, avgNumberColumn);
}
